import unittest

import shutil
import pyreindexer

# TODO Add more tests. Just checking that builtin binding is installed for sure


class TestCproto(unittest.TestCase):
    def setUp(self):
        self.db = pyreindexer.RxConnector('cproto://127.0.0.1:6534/db1')
        self.namespace = 'test_namespace'

        try:
            self.db.namespace_open(self.namespace)
        except Exception as ex:
            self.fail('open_namespace raised: ' + str(ex))

    def tearDown(self):
        self.db.namespace_close(self.namespace)
        self.db.close()

    def test_item_select(self):
        try:
            self.db.index_add(self.namespace, dict(name='id',json_paths=['id'],
                              is_pk=True,field_type='int',
                              index_type='hash'))
            self.db.item_insert (self.namespace,{'id':100, 'val':"testval"})
            ret = list (self.db.select ("SELECT * FROM " + self.namespace + " WHERE id=" + str(100)))
            assert (len(ret) == 1)
            assert (ret[0]['id'] == 100)
            assert (ret[0]['val'] == 'testval')

        except Exception as ex:
            self.fail('test_item_select raised: ' + str(ex))



from pyreindexer.raiser_mixin import RaiserMixin
from pyreindexer.query_results import QueryResults


class RxConnector(RaiserMixin):
    """ RxConnector provides a binding to Reindexer upon two shared libraries (hereinafter - APIs): 'rawpyreindexerb.so' and 'rawpyreindexerc.so'.
    The first one is aimed to a builtin way usage. That API embeds Reindexer so it could be used right in-place as is.
    The second one acts as a lightweight client which establishes a connection to Reindexer server via RPC.
    The APIs interfaces are completely the same.

    # Attributes:
        api (module): An API module loaded dynamically for Reindexer calls
        rx (int): A memory pointer to Reindexer instance
        err_code (int): the API error code
        err_msg (string): the API error message

    """

    def __init__(self, dsn):
        """Constructs a new connector object.
        Initializes an error code and a Reindexer instance descriptor to zero.

        # Arguments:
            dsn (string): The connection string which contains a protocol. Examples: 'builtin:///tmp/pyrx', 'cproto://127.0.0.1:6534/pyrx

        """

        self.err_code = 0
        self.err_msg = ''
        self.rx = 0
        self._api_import(dsn)
        self._api_init(dsn)

    def __del__(self):
        """Closes an API instance on a connector object deletion if the API is initialized

        """

        if self.rx > 0:
            self._api_close()

    def close(self):
        """Closes an API instance with Reindexer resources freeing

        # Returns:
            None

        """

        self._api_close()

    def namespace_open(self, namespace):
        """Opens a namespace specified or creates a namespace if it does not exist

        # Arguments:
            namespace (string): A name of a namespace

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.namespace_open(
            self.rx, namespace)
        self.raise_on_error()

    def namespace_close(self, namespace):
        """Closes a namespace specified

        # Arguments:
            namespace (string): A name of a namespace

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.namespace_close(
            self.rx, namespace)
        self.raise_on_error()

    def namespace_drop(self, namespace):
        """Drops a namespace specified

        # Arguments:
            namespace (string): A name of a namespace

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.namespace_drop(
            self.rx, namespace)
        self.raise_on_error()

    def namespaces_enum(self, enum_not_opened=False):
        """Gets a list of namespaces available

        # Arguments:
            enum_not_opeden (bool, optional): An enumeration mode flag. If it is
                set then closed namespaces are in result list too. Defaults to False.

        # Returns:
            (:obj:`list` of :obj:`dict`): A list of dictionaries which describe each namespace.

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg, res = self.api.namespaces_enum(
            self.rx, enum_not_opened)
        self.raise_on_error()
        return res

    def index_add(self, namespace, index_def):
        """Adds an index to the namespace specified

        # Arguments:
            namespace (string): A name of a namespace
            index_def (dict): A dictionary of index definiton

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.index_add(
            self.rx, namespace, index_def)
        self.raise_on_error()

    def index_update(self, namespace, index_def):
        """Updates an index in the namespace specified

        # Arguments:
            namespace (string): A name of a namespace
            index_def (dict): A dictionary of index definiton

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.index_update(
            self.rx, namespace, index_def)
        self.raise_on_error()

    def index_drop(self, namespace, index_name):
        """Drops an index from the namespace specified

        # Arguments:
            namespace (string): A name of a namespace
            index_name (string): A name of an index

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.index_drop(
            self.rx, namespace, index_name)
        self.raise_on_error()

    def item_insert(self, namespace, item_def, precepts=[]):
        """Inserts an item with its precepts to the namespace specified.

        # Arguments:
            namespace (string): A name of a namespace
            item_def (dict): A dictionary of item definiton
            precepts (:obj:`list` of :obj:`str`): A dictionary of index definiton

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.item_insert(
            self.rx, namespace, item_def, precepts)
        self.raise_on_error()

    def item_update(self, namespace, item_def, precepts=[]):
        """Updates an item with its precepts in the namespace specified.

        # Arguments:
            namespace (string): A name of a namespace
            item_def (dict): A dictionary of item definition
            precepts (:obj:`list` of :obj:`str`): A dictionary of index definition

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.item_update(
            self.rx, namespace, item_def, precepts)
        self.raise_on_error()

    def item_upsert(self, namespace, item_def, precepts=[]):
        """Updates an item with its precepts in the namespace specified. Creates the item if it not exist.

        # Arguments:
            namespace (string): A name of a namespace
            item_def (dict): A dictionary of item definition
            precepts (:obj:`list` of :obj:`str`): A dictionary of index definition

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.item_upsert(
            self.rx, namespace, item_def, precepts)
        self.raise_on_error()

    def item_delete(self, namespace, item_def):
        """Deletes an item from the namespace specified.

        # Arguments:
            namespace (string): A name of a namespace
            item_def (dict): A dictionary of item definition

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.item_delete(
            self.rx, namespace, item_def)
        self.raise_on_error()

    def commit(self, namespace):
        """Flushes changes to a storage of Reindexer

        # Arguments:
            namespace (string): A name of a namespace 

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.commit(
            self.rx, namespace)
        self.raise_on_error()

    def meta_put(self, namespace, key, value):
        """Puts metadata to a storage of Reindexer by key

        # Arguments:
            namespace (string): A name of a namespace
            key (string): A key in a storage of Reindexer for metadata keeping
            value (string): A metadata for storage

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.meta_put(
            self.rx, namespace, key, value)
        self.raise_on_error()

    def meta_get(self, namespace, key):
        """Gets metadata from a storage of Reindexer by key specified

        # Arguments:
            namespace (string): A name of a namespace
            key (string): A key in a storage of Reindexer where metadata is kept

        # Returns:
            string: A metadata value

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg, res = self.api.meta_get(
            self.rx, namespace, key)
        self.raise_on_error()
        return res

    def meta_enum(self, namespace):
        """Gets a list of metadata keys from a storage of Reindexer

        # Arguments:
            namespace (string): A name of a namespace

        # Returns:
            (:obj:`list` of :obj:`str`): A list of all metadata keys.

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg, res = self.api.meta_enum(
            self.rx, namespace)
        self.raise_on_error()
        return res

    def select(self, query):
        """Executes an SQL query and returns query results

        # Arguments:
            query (string): An SQL query

        # Returns:
            (:obj:`QueryResults`): A QueryResults iterator.

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.raise_on_not_init()
        self.err_code, self.err_msg, qres_wrapper_ptr, qres_iter_count = self.api.select(
            self.rx, query)
        self.raise_on_error()
        return QueryResults(self.api, qres_wrapper_ptr, qres_iter_count)

    def _api_import(self, dsn):
        """Imports an API dynamically depending on protocol specified in dsn.

        # Arguments:
            dsn (string): The connection string which contains a protocol.

        # Returns:
            None

        # Raises:
            Exception: Raises an exception if a connection protocol is unrecognized.

        """

        if dsn.startswith('builtin://'):
            self.api = __import__('rawpyreindexerb')
        elif dsn.startswith('cproto://'):
            self.api = __import__('rawpyreindexerc')
        else:
            raise Exception(
                "Unknown Reindexer connection protocol for dsn: ", dsn)

    def _api_init(self, dsn):
        """Initializes Reindexer instance and connects to a database specified in dsn.
        Obtains a pointer to Reindexer instance.

        # Arguments:
            dsn (string): The connection string which contains a protocol.

        # Returns:
            None

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.
            Exception: Raises with an error message of API return on non-zero error code.

        """

        self.rx = self.api.init()
        self.raise_on_not_init()
        self.err_code, self.err_msg = self.api.connect(self.rx, dsn)
        self.raise_on_error()

    def _api_close(self):
        """Destructs Reindexer instance correctly and resets memory pointer.

        # Returns:
            None

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.

        """

        self.raise_on_not_init()
        self.api.destroy(self.rx)
        self.rx = 0



from pyreindexer import RxConnector


def create_index_example(db, namespace):
    index_definition = {
        'name': 'id',
        'json_paths': ['id'],
        'field_type': 'int',
        'index_type': 'hash',
        'is_pk': True,
        'is_array': False,
        'is_dense': False,
        'is_sparse': False,
        'collate_mode': 'none',
        'sort_order_letters': '',
        'expire_after':0,
        'config': {},
    }

    try:
        db.index_add(namespace, index_definition)
    except Exception:
        db.index_drop(namespace, 'id')
        db.index_add(namespace, index_definition)


def update_index_example(db, namespace):
    index_definition_modified = {
        'name': 'id',
        'json_paths': ['id'],
        'field_type': 'int64',
        'index_type': 'hash',
        'is_pk': True,
        'is_array': False,
        'is_dense': True,
        'is_sparse': False,
        'collate_mode': 'none',
        'sort_order_letters': '',
        'expire_after':0,
        'config': {},
    }
    db.index_update(namespace, index_definition_modified)


def create_items_example(db, namespace):
    ITEMS_COUNT = 10

    for i in range(0, ITEMS_COUNT):
        item = {'id': i + 1, 'name': 'item_' + str(i % 2)}
        db.item_upsert(namespace, item)


def select_item_query_example(db, namespace):
    item_name_for_lookup = 'item_0'

    return db.select('SELECT * FROM ' + namespace + ' WHERE name="' + item_name_for_lookup + '"')


def rx_example():
    db = RxConnector('builtin:///tmp/pyrx')
    #db = RxConnector('cproto://127.0.0.1:6534/pyrx')

    namespace = 'test_table'

    db.namespace_open(namespace)

    create_index_example(db, namespace)
    update_index_example(db, namespace)

    create_items_example(db, namespace)

    selected_items = select_item_query_example(db, namespace)

    res_count = selected_items.count()
    print('Results count: ', res_count)

    # disposable QueryResults iterator
    for item in selected_items:
        print('Item: ', item)

    # won't be iterated again
    for item in selected_items:
        print('Item: ', item)

    db.close()


if __name__ == "__main__":
    rx_example()



import unittest

import shutil
import pyreindexer

# TODO Add more tests. Just checking that builtin binding is installed for sure


class TestBuiltin(unittest.TestCase):
    def setUp(self):
        self.db_path = '/tmp/pyreindexer_builtin_test_db'
        self.namespace = 'test_namespace'

        self.db = pyreindexer.RxConnector('builtin://' + self.db_path)

        try:
            self.db.namespace_open(self.namespace)
        except Exception as ex:
            self.fail('open_namespace raised: ' + str(ex))

    def tearDown(self):
        self.db.namespace_close(self.namespace)
        self.db.close()

        shutil.rmtree(self.db_path, ignore_errors=True)

    def test_index_create(self):
        index_definition = dict(name='id', json_paths=['id'], field_type='int',
                                index_type='hash', is_pk=True, is_array=False,
                                is_dense=False, is_sparse=False, collate_mode='none',
                                sort_order_letters='', config={})
        try:
            self.db.index_add(self.namespace, index_definition)
        except Exception as ex:
            self.fail('index_create raised: ' + str(ex))

        index_definition['field_type'] = 'int64'

        with self.assertRaises(Exception):
            self.db.index_add(self.namespace, index_definition)

    def test_item_insert(self):
        try:
            self.db.index_add(self.namespace, dict(name='id',json_paths=['id'],
                              is_pk=True,field_type='int',
                              index_type='hash'))
            self.db.item_insert (self.namespace,{'id':100, 'val':"testval"})
        except Exception as ex:
            self.fail('item_insert raised: ' + str(ex))

    def test_item_select(self):
        try:
            self.db.index_add(self.namespace, dict(name='id',json_paths=['id'],
                              is_pk=True,field_type='int',
                              index_type='hash'))
            self.db.item_insert (self.namespace,{'id':100, 'val':"testval"})
            ret = list (self.db.select ("SELECT * FROM " + self.namespace + " WHERE id=" + str(100)))
            assert (len(ret) == 1)
            assert (ret[0]['id'] == 100)
            assert (ret[0]['val'] == 'testval')

        except Exception as ex:
            self.fail('test_item_select raised: ' + str(ex))



"""
The pyreindexer module provides a connector and its auxiliary tools for interaction with Reindexer.
"""

from pyreindexer.rx_connector import RxConnector
# from pyreindexer.index_definition import IndexDefinition



import os
import sys
from multiprocessing import cpu_count

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as build_ext_orig

if sys.version_info < (3, 6):
    raise RuntimeError('Require Python 3.6 or greater')

PACKAGE_NAME = 'pyreindexer'

class CMakeExtension(Extension):
    def __init__(self, name):
        super().__init__(name, sources=[])


class build_ext(build_ext_orig):
    def run(self):
        for ext in self.extensions:
            self.build_cmake(ext)

    def build_cmake(self, ext):
        cwd = os.path.abspath('')

        build_temp = os.path.abspath(self.build_temp)
        if not os.path.exists(build_temp):
            os.makedirs(build_temp)

        extension_dir = os.path.abspath(self.get_ext_fullpath(ext.name))
        if not os.path.exists(extension_dir):
            os.makedirs(extension_dir)

        os.chdir(build_temp)

        lib_dir = os.path.join(extension_dir, '..')
        source_dir = os.path.join(cwd, PACKAGE_NAME)

        self.spawn(['cmake', source_dir,
                    '-DCMAKE_BUILD_TYPE=Release',
                    '-DCMAKE_CXX_STANDARD=17',
                    '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + lib_dir])
        if not self.dry_run:
            self.spawn(['cmake', '--build', '.'])

        os.chdir(cwd)


setup(name=PACKAGE_NAME,
      version='0.2.1',
      description='A connector that allows to interact with Reindexer',
      url='https://github.com/Restream/pyreindexer',
      author='Igor Tulmentyev',
      author_email='igtulm@gmail.com',
      maintainer='Oleg Gerasimov',
      maintainer_email='ogerasimov@gmail.com',
      license='Apache License 2.0',
      packages=[PACKAGE_NAME],
      ext_modules=[CMakeExtension('rawpyreindexer')],
      cmdclass={'build_ext': build_ext},
      package_data={'pyreindexer': [
          'CMakeLists.txt',
          'lib/include/pyobjtools.h',
          'lib/include/pyobjtools.cc',
          'lib/include/queryresults_wrapper.h',
          'lib/src/rawpyreindexer.h',
          'lib/src/rawpyreindexer.cc',
          'lib/src/reindexerinterface.h',
          'lib/src/reindexerinterface.cc',
          'example/main.py',
          'tests/__init__.py',
          'tests/test_builtin.py',
          'tests/test_cproto.py'
      ]},
      test_suite='tests', install_requires=['envoy', 'delegator', 'pytest']
      )



class RaiserMixin(object):
    """ RaiserMixin contains methods for checking some typical API bad events and raise if there is a necessity.

    """

    def raise_on_error(self):
        """Checks if there is an error code and raises with an error message.

        # Raises:
            Exception: Raises with an error message of API return on non-zero error code.

        """

        if self.err_code:
            raise Exception(self.err_msg)

    def raise_on_not_init(self):
        """Checks if there is an error code and raises with an error message.

        # Raises:
            Exception: Raises with an error message of API return if Reindexer instance is not initialized yet.

        """

        if self.rx <= 0:
            raise Exception("Connection is not initialized")
